// **************************************************************************
// **** ZX81 USB Keyboard for Funtronics LeoStick (based on a Leonardo). ****
// **** zx81usbkeyboard.ino                                              ****
// **************************************************************************
// ** David Stephenson 2016-08-08  **
// **                              **
// ** Originally based on code by: **
// ** Dave Curran  2013-04-27      **
// ** Tony Smith 2014-02-15        **
// **********************************

#include "Arduino.h"
#include "zx81keyboard.h"
#include "zx81modestate.h"
 
// ************************
// *** Global Variables ***
// ************************

// Setup Global Variables for keyboard.
ZxKeyBoard MyKeyboard;

// Setup LeoStick pins for mode switch and indicator LEDs.
ModeState MyModeState(A5,A0,A1,A2);

// Setup LeoStick pins for Keyboard rows and columns.
const byte bColPins[NUM_COLS] = {13,12,10,9,8};
const byte bRowPins[NUM_ROWS] = {7,6,5,4,3,2,1,0};

// Pin that piezo is connected to. should be 11 on a LeoStick
const byte bBeepPin = 11;

// boDebEmu off by default. Turn on via serial port commands.
// Set STANDARD mode keyboard debounce to EMULATOR timings if true.
// Good to set if switching to standard mode while in an EMU frequently.
// Set to false to speed up type-o-matic repeat rate out off EMU mode
boolean boDebEmu = false;

// Beep on keypress off by default. Turn on via serial port commands.
// Beep always off in Emulator mode.
// boDebEmu must be set to false and boKeyBeep must be true for beep to function.
boolean boKeyBeep = false;

// Serial Port in, for reading commands.
String sSerialIn = "";

// ******************
// *** Main Setup ***
// ******************
void setup() {
	 
	// Set all Keyboard col pins as inputs and activate pull-ups.
	for (byte bColCount = 0 ; bColCount < NUM_COLS ; bColCount++)
	{
		pinMode(bColPins[bColCount], INPUT);
		digitalWrite(bColPins[bColCount], HIGH);
	}
	
	// Set all Keyboard row pins as inputs.
	for (byte bRowCount = 0 ; bRowCount < NUM_ROWS ; bRowCount++)
	{
		pinMode(bRowPins[bRowCount], INPUT);
	}
	
	// initialize control over the keyboard.
	Keyboard.begin();
	// initialize serial port
	Serial.begin(115200);
	
}
 
// ************************
// *** Main loop        ***
// *** Lets get Busy-ah ***
//*************************
void loop() {

	bool boShifted = false;
	byte bKeyPressed = 0;
	
	// Set Keyboard Mode.
	KEYMODES eMyKeyMode = MyModeState.GetMode();
	KEYSTATES eMyKeyState = MyModeState.GetState();
	
	// Check for the Shift key being pressed.
	pinMode(bRowPins[SHIFT_ROW], OUTPUT);
	
	if (digitalRead(bColPins[SHIFT_COL]) == LOW) boShifted = true;
	
	pinMode(bRowPins[SHIFT_ROW], INPUT);
	
	for (byte bRow = 0 ; bRow < NUM_ROWS ; bRow++){
		// Run through the rows, turn them on.
		
		pinMode(bRowPins[bRow], OUTPUT);
		digitalWrite(bRowPins[bRow], LOW);
		
		for (byte bCol = 0 ; bCol < NUM_COLS ; bCol++){
			if (digitalRead(bColPins[bCol]) == LOW){
				if (boShifted && eMyKeyMode == STANDARD){
					// Select correct Keyboard layout for current state For STANDARD mode. 
					// Shift alters the Current state selection. eg. gets Red Shift symbols in Normal State. 
					switch (eMyKeyState) {
						case NORMAL:
							bKeyPressed = MyKeyboard.bKeyPress(bRow, bCol, NORMAL_SHIFTED, eMyKeyMode);
							if (bKeyPressed == KEY_RETURN) { 
								eMyKeyState = FUNCTION;
								bKeyPressed = 0;
								MyKeyboard.KeyDisable(bRow, bCol);
							}
							if (bKeyPressed == '9') {
								eMyKeyState = GRAPHICS;
								bKeyPressed = 0;
								MyKeyboard.KeyDisable(bRow, bCol);
							}
						break;
							
						case FUNCTION:
							bKeyPressed = MyKeyboard.bKeyPress(bRow, bCol, NORMAL_SHIFTED, eMyKeyMode);
							if (bKeyPressed == KEY_RETURN) { 
								eMyKeyState = NORMAL;
								bKeyPressed = 0;
								MyKeyboard.KeyDisable(bRow, bCol);
							}
							if (bKeyPressed == '9') {
								eMyKeyState = GRAPHICS;
								bKeyPressed = 0;
								MyKeyboard.KeyDisable(bRow, bCol);
							}
						break;
							
						case GRAPHICS:
							bKeyPressed = MyKeyboard.bKeyPress(bRow, bCol, GRAPHICS_SHIFTED, eMyKeyMode);
							if (bKeyPressed == KEY_RETURN) { 
								eMyKeyState = FUNCTION;
								bKeyPressed = 0;
								MyKeyboard.KeyDisable(bRow, bCol);
							}
							if (bKeyPressed == '9') {
								eMyKeyState = NORMAL;
								bKeyPressed = 0;
								MyKeyboard.KeyDisable(bRow, bCol);
							}
						break;
							
					}
				} else {
					// Emulator keyboard, the keyboard states are controlled by a ZX81 emulator.
					// Keyboard mimics unaltered PS2 Keyboard presses as expected by an emulator.
					bKeyPressed = MyKeyboard.bKeyPress(bRow, bCol, eMyKeyState, eMyKeyMode);
				}
				
				if (bKeyPressed > 0 ) {
					//Serial.write(bKeyPressed);
					if (eMyKeyMode == EMULATOR){
						if (bKeyPressed && boShifted) Keyboard.press(KEY_LEFT_SHIFT);
						Keyboard.press(bKeyPressed);
						// Many Emulators don't use standard OS keyboard routines and require a pause
						// for key to register. There is some variance, but 100ms seems to cover most.  
						delay(DEBOUNCE_DELAY/2);
					} else {
						if (eMyKeyState == GRAPHICS && bKeyPressed > 96 && bKeyPressed < 123){
							if (boShifted){
								Keyboard.press(KEY_LEFT_ALT);
							} else {
								Keyboard.press(KEY_LEFT_CTRL);
							}
						}
						Keyboard.press(bKeyPressed);
						// Some audio feedback if not in EMULATOR mode.
                                                if (boDebEmu == false){
                                                        if (boKeyBeep == true) tone(bBeepPin, 31, 20);
                                                } else {
					                delay(DEBOUNCE_DELAY/2);
						}
					}
					
					Keyboard.releaseAll();
				}
				
			} else {
				MyKeyboard.KeyPressReset(bRow, bCol);
			}
		}
		pinMode(bRowPins[bRow], INPUT);
	}
	
	digitalWrite(bRowPins[SHIFT_ROW], LOW);
	
	// Check for Commands / updates from USB Serial port
	if (Serial.available() > 0) commandIn();
	
	// Update LED panel and check for Mode change switch press.
	MyModeState.SetState(eMyKeyState);
	
 }
 
 void commandIn(){
	 
	 //sSerialIn.reserve(11);
	 
	 //while (Serial.available()) {
	 char inChar = (char)Serial.read();
	 sSerialIn += inChar;
	 sSerialIn.trim();
	 sSerialIn.toUpperCase();
	 
	 String KEYWORDS[8] = {"DEVICE","EMULATOR","STANDARD","BELL","BEEPON","BEEPOFF","DBEMON","DBEMOFF"};
	 
	 if (sSerialIn.length() > 8) sSerialIn = sSerialIn.substring(1,9);
	 for (byte bKeyword = 0 ; bKeyword < 8 ; bKeyword++){
		 if (sSerialIn.indexOf(KEYWORDS[bKeyword])!=-1){
			 if (KEYWORDS[bKeyword] == "DEVICE"){
				 Serial.println("ZX81_KEYS");
			 }
			 else if (KEYWORDS[bKeyword] == "BELL"){
				 tone(bBeepPin, 262, 60);
				 tone(bBeepPin, 196, 40);
			}
					 
			else if (KEYWORDS[bKeyword] == "EMULATOR"){ 
				 MyModeState.SetMode(EMULATOR);
			}
			
			else if (KEYWORDS[bKeyword] == "STANDARD"){
				MyModeState.SetMode(STANDARD);
			}

			else if (KEYWORDS[bKeyword] == "DBEMON"){
                                boDebEmu = true;
                        }
			 
			else if (KEYWORDS[bKeyword] =="DBEMOFF"){
                                boDebEmu = false;
                        }

			else if (KEYWORDS[bKeyword] == "BEEPON"){
			        boKeyBeep = true;
			}
			 
			else if (KEYWORDS[bKeyword] == "BEEPOFF"){
				 boKeyBeep = false;
			}
			 
			// Clear Value if String found.
			sSerialIn = "";
		 }
	 }
	 
 }
 
